#!/usr/bin/perl
#
# fetch_random_image.pl
#
#	ImagesForever 1.9
#	by Scott Lahteine
#   Thinkyhead Software
#	December 10, 2011
#
#	pass:	[0] serial id for a fetch request
#       [1] a flag to do safe search (1) or not (0)
#       [2] image sizes to fetch
#       [3] the path to the banfile
#       [4] escaped search terms
#
#   or:     -1 thru -20 to get an image from the history
#
#	returns:    space-delimited results:
#			 1. serial number
#			 2. try count
#			 3. total found
#			 4. final term
#			 5. safesearch pref
#			 6. image size pref
#			 7. results page
#			 8. original image url
#            9. a page the image appears on
#			10. original image width
#			11. original image height
#			12. the cached image file
#

print('call error'), exit if ($#ARGV < 0 || $#ARGV > 4);

my $histsize = 100;

my $homedir = `cd ~; pwd`; chop $homedir;
my $tmp_dir	= $homedir . "/Library/Application Support/ImagesForever/Cache";
my $esc_tmp = $tmp_dir; $esc_tmp =~ s/ /\\ /;
my $histfile = "$tmp_dir/.history86";
my $esc_hist = $histfile; $esc_hist =~ s/ /\\ /;
my $agent	= "'Mozilla/5.001 (windows; U; NT4.0; en-us) Gecko/25250101'";
my $curl	= "curl -f -s -m 120 -A $agent";
my $prefix	= 'img-';
my $googles	= 0;
my $max_tries = 10;
my $total_results = 0;
my $query;

`mkdir -p $esc_tmp`;

# get the history or a new image
if ($ARGV[0] =~ /-\d+/)
{
	fetch_history($ARGV[0]);
}
else
{
	#
	# Get the request serial #
	#
	$serial = $ARGV[0];

	#
	# Make the query seed
	#
	@sizes = ('','xxlarge','small|medium|large|xlarge','icon');

	$query = 'http://images.google.com/images?filter=0'
				.($ARGV[1]?'':'&safe=off')
				.($ARGV[2]?"&imgsz=$sizes[$ARGV[2]]":'')
				.'&q=';

	#
	# Get the list of user-blocked terms and urls
	#
	if (open(FILE, "<$ARGV[3]")) {
		chop, $banned{$_}=1 while <FILE>;
		close FILE;
	}

	fetch_random_image($ARGV[4]);
}

#------------------------
# add_to_history
#
# pass:		the output from fetch_random_image
# returns:	nothing
#
sub add_to_history
{
	$cnt = $histsize - 1;
	`touch $esc_hist`;
	`tail -$cnt $esc_hist >$esc_hist~`;
	`mv $esc_hist~ $esc_hist`;
	open (FILE, ">>$histfile");
	print FILE $_[0];
	close FILE;

	clean_images();
}

#------------------------
# fetch_history
#
# pass:		-1 thru -20, the position to fetch
# returns:	a line just like fetch_random_image
#
sub fetch_history
{
	$num = $_[0];

	print('history error'), exit if !open(FILE, "<$histfile");
	chop, push @hist, ($_) while (<FILE>);
	close FILE;
    $out = $hist[$num];
    @arr = split(/ /, $out);
    if ($#arr < 11) {
    	@bits = split(/\//, $arr[7]);
    	splice @arr, 8, 0, $bits[0];
    	$out = "@arr";
    }
    print $out;
}


#------------------------
# clean_images
#
# cleans out older images
# leaves a few extra besides
# the history buffer (20)
#
# pass:		nothing
# returns:	nothing
#
sub clean_images
{
	@list = `ls -1 $esc_tmp | sort`;
	$num = $#list - $histsize;
	`rm $esc_tmp/$list[$num]` while $num-- > 0;
}

#------------------------
# fetch_random_image
#
# pass:		escaped search string
# returns:	prints the results as a space-delimited string
#
sub fetch_random_image
{
	$word = shift;

	opendir(DIR, $tmp_dir) ? closedir DIR : mkdir($tmp_dir);

FETCH: while(1)
	{
		$word = get_random_word() unless $word;

		print('dictionary error'), exit unless $word;

		$scrape = google_image_search($word);
		$len = length($scrape);

		print('search request error'), exit unless $len;

		last FETCH if (++$googles > $max_tries);

		$word = '', next if ($len < 3500);

		@page_list = extract_page_list($scrape);

		$page_try = $pages = $#page_list + 1;
		$other = $y = int(rand($pages));

		while($page_try--)
		{
			$pp = $page_list[$y++ % $pages];

			if ($other++) {
				$scrape = google_image_search($word, $pp);
				last FETCH if (++$googles > $max_tries);
			}

			@img_list = extract_image_list($scrape);
			$img_try = $images = $#img_list + 1;

			print('results scrape error'), exit unless $images;

			$x = int(rand($images));

			while ($img_try--) {
				$desc = $img_list[$x++ % $images];
				($url, $ref, $w, $h) = split(/ /, $desc);
				last FETCH if $fn=fetch_image_from_url($url);
			}
		}

		$word = '';
	}

	print ('unable to google'), exit if $googles > $max_tries;

	$desc =~ s!http://!!i;
	$out = "$googles $total_results $word $ARGV[1] $ARGV[2] $pp $desc $fn";
	add_to_history("x $out\n");

	print "$serial $out";
}

#------------------------
# fetch_image_from_url
#
# pass:		the url to fetch
# returns:	the file name or an empty string
# todo:		add dead urls to the blocked list
#
sub fetch_image_from_url
{
	$_[0] =~ /(\.[^.]+)$/;
	$ext = $1;
	$fna = $prefix.time;
	$fn = $fna.$ext;

	$url = "http://$url" if $url !~ /http:\/\//;

	`$curl -o $esc_tmp/$fn "$url"`;
	$type = stat("$tmp_dir/$fn") ? `file $esc_tmp/$fn` : '';
	$fn = (($type =~ /(JPEG|GIF|PNG) image data/s) ? $fn : '');

	# rename if the real type doesn't match the .ext
	if ($fn) {
		$type = lc $1;
		$test = "$type$ext";
		if ($test !~ /jpeg\.jpe?g|gif\.gif|png\.png/i) {
			$fn2 = "$fna.$type";
			rename("$tmp_dir/$fn","$tmp_dir/$fn2");
			$fn = $fn2;
		}
	}

	$fn;
}

#------------------------
# google_image_search
#
# pass:		escaped search string
#			[first image number]
# returns:	a page of html-formatted search results
#
sub google_image_search
{
	my ($word, $start) = @_;

	$word .= "&start=$start" if $start;
	$search = "$curl \"$query$word\"";
	`$search`;
}

#------------------------
# extract_image_list
#
# pass:		google results html page
# returns:	array of image statistics
#
sub extract_image_list
{
	my $scrape = shift;
	my @img_list;

	#
	# 2011-6-1
    #
    # Google results now have named arguments in a url form
	#
	# This function extracts arguments from the july'11 url style
	#

	$x = 0;
	$scrape =~ s/\\x([A-Fa-f0-9]{2})/pack('C', hex($1))/seg;
	while ($scrape =~ /"\/imgres\?[^"]*imgurl=([^&]*)&[^"]*imgrefurl=([^&]*)&[^"]*h=(\d+)&[^"]*w=(\d+)[^"]*"/isg) {
		($url, $ref, $h, $w) = ($1, $2, $3, $4);
		next if exists $banned{$url};		# throw away blocked urls
		$img_list[$x++] = "$url $ref $w $h";
	}

	@img_list;
}

#------------------------
# extract_page_list
#
# pass:		google results html page
# returns:	array of secondary results pages
#
sub extract_page_list
{
	my $scrape = shift;
	my @page_list;

#
# Google results looks like this:
#
# <div>Results <b>1</b> - <b>21</b> of about <b>590,000</b> for <b>crustaceans</b>.</div>
#

	if ($scrape =~ m|<div>Results <b>\d+</b> - <b>\d+</b> of[ a-z]+<b>([\d,]+)</b>|isg) {
		$total = $1;
		$total =~ s/[^\d]//g;
		$total_results = $total;
		$total = 980 if $total > 980;
		$x=-20;
		push @page_list, ($x) while ($x+=20) < $total;
	}

	@page_list;
}

#------------------------
# get_random_word
#
# pass:		flag to include the web2a dictionary
# returns:	a random word
#
sub get_random_word
{
	$dict = '/usr/share/dict/' . ($_[0] && rand(234937+76205)<76205 ? 'web2a' : 'web2');

	$size = (stat($dict))[7];
	$word = '';

	while(!$word) {
		open(FILE, "<$dict");
		seek(FILE, rand($size-60), SEEK_SET);
		read(FILE, $sample, 60);
		close FILE;

		@words = split(/\n\r|\n|\r/, $sample);
		$word = $words[rand($#words-1)+1];
		$word = '' if $word !~ /..../ || exists $banned{$word};
	}

	$word =~ s/ /+/g;
	$word;
}
